#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include "StackFunc.h"
#include "AkelDLL.h"


//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackDelete
#define StackClear
#include "StackFunc.h"

//Defines
#ifndef EM_SHOWSCROLLBAR
  #define EM_SHOWSCROLLBAR (WM_USER + 96)
#endif

typedef struct _FRAMEWINDOW {
  struct _FRAMEWINDOW *next;
  struct _FRAMEWINDOW *prev;
  HWND hWnd;
  RECT rc;
  DWORD dwStyle;
} FRAMEWINDOW;

//Functions prototypes
void CALLBACK NewFrameProcRet(CWPRETSTRUCT *cwprs);
FRAMEWINDOW* StackInsertFrameWindow(HSTACK *hStack);
void StackDeleteFrameWindow(HSTACK *hStack, HWND hWnd);
void StackGetRectFrameWindow(HSTACK *hStack);
void StackHideFrameWindow(HSTACK *hStack);
void StackShowFrameWindow(HSTACK *hStack);
void StackFreeFrameWindow(HSTACK *hStack);
void HideFrameWindow(HWND hWnd, int x, int y, int cx, int cy);
void ShowFrameWindow(HWND hWnd, int x, int y, int cx, int cy);
void ShowEditScrollBars(HWND hWnd, BOOL bShow);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
HWND hMainWnd;
HWND hWndEdit;
HWND hStatus;
HWND hMdiClient;
HWND hTab;
BOOL bOldRichEdit;
BOOL bMDI;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
HMENU hMainMenu;
int nScreenX=0;
int nScreenY=0;
RECT rcMain;
BOOL bStatusVisible;
BOOL bTabVisible;
BOOL bFrameMaximized;
BOOL bHideEditScrollBars=FALSE;
HSTACK hFrameWindowStack={0};
WNDPROCRETDATA *NewFrameProcRetData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="FullScreen";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  if (pd->bOnStart) return;
  if (!bInitCommon) InitCommon(pd);

  if (bInitMain)
  {
    UninitMain();
  }
  else
  {
    InitMain();

    //Stay in memory, and show as active
    pd->nUnload=UD_NONUNLOAD_ACTIVE;
  }
}

void CALLBACK NewFrameProcRet(CWPRETSTRUCT *cwprs)
{
  if (cwprs->message == WM_MDIACTIVATE)
  {
    if (cwprs->hwnd == (HWND)cwprs->lParam)
    {
      FRAMEWINDOW *tmp;
      DWORD dwStyle;

      dwStyle=GetWindowLongA(cwprs->hwnd, GWL_STYLE);

      if (dwStyle & WS_OVERLAPPEDWINDOW)
      {
        if (tmp=StackInsertFrameWindow(&hFrameWindowStack))
        {
          tmp->hWnd=cwprs->hwnd;
          GetWindowRect(cwprs->hwnd, &tmp->rc);
          tmp->rc.right-=tmp->rc.left;
          tmp->rc.bottom-=tmp->rc.top;
          ScreenToClient(hMdiClient, (POINT *)&tmp->rc.left);
        }
        HideFrameWindow(cwprs->hwnd, 0, 0, nScreenX, nScreenY);
      }
      SetFocus(cwprs->hwnd);
    }
  }
  else if (cwprs->message == WM_DESTROY)
  {
    StackDeleteFrameWindow(&hFrameWindowStack, cwprs->hwnd);
  }

  //Call next procedure
  if (NewFrameProcRetData->NextProc)
    NewFrameProcRetData->NextProc(cwprs);
}

FRAMEWINDOW* StackInsertFrameWindow(HSTACK *hStack)
{
  FRAMEWINDOW *element=NULL;

  StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&element, -1, sizeof(FRAMEWINDOW));
  return element;
}

void StackDeleteFrameWindow(HSTACK *hStack, HWND hWnd)
{
  FRAMEWINDOW *tmp=(FRAMEWINDOW *)hStack->first;

  while (tmp)
  {
    if (tmp->hWnd == hWnd)
    {
      StackDelete((stack **)&hStack->first, (stack **)&hStack->last, (stack *)tmp);
      return;
    }
    tmp=tmp->next;
  }
}

void StackGetRectFrameWindow(HSTACK *hStack)
{
  FRAMEWINDOW *tmp=(FRAMEWINDOW *)hStack->first;

  while (tmp)
  {
    tmp->dwStyle=GetWindowLongA(tmp->hWnd, GWL_STYLE);
    if (tmp->dwStyle & WS_MINIMIZE)
      ShowWindow(tmp->hWnd, SW_SHOWNORMAL);
    GetWindowRect(tmp->hWnd, &tmp->rc);
    tmp->rc.right-=tmp->rc.left;
    tmp->rc.bottom-=tmp->rc.top;
    ScreenToClient(hMdiClient, (POINT *)&tmp->rc.left);

    tmp=tmp->next;
  }
}

void StackHideFrameWindow(HSTACK *hStack)
{
  FRAMEWINDOW *tmp=(FRAMEWINDOW *)hStack->first;

  while (tmp)
  {
    HideFrameWindow(tmp->hWnd, 0, 0, nScreenX, nScreenY);

    tmp=tmp->next;
  }
}

void StackShowFrameWindow(HSTACK *hStack)
{
  FRAMEWINDOW *tmp=(FRAMEWINDOW *)hStack->first;

  while (tmp)
  {
    ShowFrameWindow(tmp->hWnd, tmp->rc.left, tmp->rc.top, tmp->rc.right, tmp->rc.bottom);
    if (tmp->dwStyle & WS_MINIMIZE)
      ShowWindow(tmp->hWnd, SW_MINIMIZE);

    tmp=tmp->next;
  }
}

void StackFreeFrameWindow(HSTACK *hStack)
{
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void HideFrameWindow(HWND hWnd, int x, int y, int cx, int cy)
{
  HWND hWndFrameEdit;
  DWORD dwStyle;

  dwStyle=GetWindowLongA(hWnd, GWL_STYLE);
  SetWindowLongA(hWnd, GWL_STYLE, dwStyle & ~WS_OVERLAPPEDWINDOW);
  MoveWindow(hWnd, x, y, cx, cy, TRUE);

  if (bHideEditScrollBars)
  {
    if (!bOldRichEdit)
    {
      hWndFrameEdit=GetDlgItem(hWnd, ID_EDIT);
      ShowEditScrollBars(hWndFrameEdit, FALSE);
    }
  }
}

void ShowFrameWindow(HWND hWnd, int x, int y, int cx, int cy)
{
  HWND hWndFrameEdit;
  DWORD dwStyle;

  dwStyle=GetWindowLongA(hWnd, GWL_STYLE);
  SetWindowLongA(hWnd, GWL_STYLE, dwStyle | WS_OVERLAPPEDWINDOW);
  MoveWindow(hWnd, x, y, cx, cy, TRUE);

  if (bHideEditScrollBars)
  {
    if (!bOldRichEdit)
    {
      hWndFrameEdit=GetDlgItem(hWnd, ID_EDIT);
      ShowEditScrollBars(hWndFrameEdit, TRUE);
    }
  }
}

void ShowEditScrollBars(HWND hWnd, BOOL bShow)
{
  EDITINFO ei;

  if (!bOldRichEdit)
  {
    if (bShow)
    {
      //Vertical
      SendMessage(hWnd, EM_SHOWSCROLLBAR, SB_VERT, TRUE);

      //Horizontal
      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)&ei))
      {
        if (!ei.bWordWrap)
          SendMessage(hWnd, EM_SHOWSCROLLBAR, SB_HORZ, TRUE);
      }
    }
    else
    {
      //Vertical
      SendMessage(hWnd, EM_SHOWSCROLLBAR, SB_VERT, FALSE);

      //Horizontal
      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)&ei))
      {
        if (!ei.bWordWrap)
          SendMessage(hWnd, EM_SHOWSCROLLBAR, SB_HORZ, FALSE);
      }
    }
  }
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hMainWnd=pd->hMainWnd;
  hWndEdit=pd->hWndEdit;
  hStatus=pd->hStatus;
  hMdiClient=pd->hMdiClient;
  hTab=pd->hTab;
  bOldRichEdit=pd->bOldRichEdit;
  bMDI=pd->bMDI;
}

void InitMain()
{
  FRAMEWINDOW *tmp;
  HWND hWndFrameActive;
  HWND hWndChild;
  DWORD dwStyle;

  bInitMain=TRUE;

  //Screen metrics
  nScreenX=GetSystemMetrics(SM_CXSCREEN);
  nScreenY=GetSystemMetrics(SM_CYSCREEN);

  //Change main window caption and size
  dwStyle=GetWindowLongA(hMainWnd, GWL_STYLE);
  if (!(dwStyle & WS_MAXIMIZE))
  {
    GetWindowRect(hMainWnd, &rcMain);
    rcMain.right-=rcMain.left;
    rcMain.bottom-=rcMain.top;
  }
  SetWindowLongA(hMainWnd, GWL_STYLE, (dwStyle & ~WS_OVERLAPPEDWINDOW) | WS_POPUP);
  MoveWindow(hMainWnd, 0, 0, nScreenX, nScreenY, TRUE);

  if (!bMDI)
  {
    //Hide main window menu
    hMainMenu=GetMenu(hMainWnd);
    SetMenu(hMainWnd, NULL);

    //Hide main window children
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (hWndChild != hWndEdit)
        ShowWindow(hWndChild, SW_HIDE);
      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

    //Change edit window size
    MoveWindow(hWndEdit, 0, 0, nScreenX, nScreenY, TRUE);

    //Hide edit window scrollbars
    if (bHideEditScrollBars)
    {
      ShowEditScrollBars(hWndEdit, FALSE);
    }
  }
  else
  {
    //Remember MDI frame windows
    hWndChild=GetWindow(hMdiClient, GW_CHILD);

    while (hWndChild)
    {
      if (tmp=StackInsertFrameWindow(&hFrameWindowStack))
        tmp->hWnd=hWndChild;
      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

    //Remember MDI frame sizes
    SendMessage(hMdiClient, WM_SETREDRAW, FALSE, 0);
    if (hWndFrameActive=(HWND)SendMessage(hMdiClient, WM_MDIGETACTIVE, 0, (LPARAM)&bFrameMaximized))
    {
      if (bFrameMaximized) SendMessage(hMdiClient, WM_MDIRESTORE, (WPARAM)hWndFrameActive, 0);
      StackGetRectFrameWindow(&hFrameWindowStack);
      SetFocus(hWndFrameActive);
    }
    SendMessage(hMdiClient, WM_SETREDRAW, TRUE, 0);

    //Hide main window menu
    hMainMenu=GetMenu(hMainWnd);
    SetMenu(hMainWnd, NULL);

    //Hide main window children
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (hWndChild == hStatus)
      {
        if (bStatusVisible=IsWindowVisible(hStatus))
          ShowWindow(hStatus, SW_HIDE);
      }
      else if (hWndChild == hTab)
      {
        if (bTabVisible=IsWindowVisible(hTab))
          ShowWindow(hTab, SW_HIDE);
      }
      else if (hWndChild != hMdiClient)
        ShowWindow(hWndChild, SW_HIDE);

      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

    //Change MDI client window size
    MoveWindow(hMdiClient, 0, 0, nScreenX, nScreenY, TRUE);

    //Resize MDI frame windows
    SendMessage(hMdiClient, WM_SETREDRAW, FALSE, 0);
    StackHideFrameWindow(&hFrameWindowStack);
    SendMessage(hMdiClient, WM_SETREDRAW, TRUE, 0);
    InvalidateRect(hWndFrameActive, NULL, FALSE);

    //SubClass
    NewFrameProcRetData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROCRET, (WPARAM)NewFrameProcRet, (LPARAM)&NewFrameProcRetData);
  }
}

void UninitMain()
{
  HWND hWndFrameActive;
  HWND hWndChild;
  DWORD dwStyle;

  bInitMain=FALSE;

  //Restore main window caption and size
  dwStyle=GetWindowLongA(hMainWnd, GWL_STYLE);
  SetWindowLongA(hMainWnd, GWL_STYLE, (dwStyle & ~WS_POPUP) | WS_OVERLAPPEDWINDOW);
  if (!(dwStyle & WS_MAXIMIZE))
    MoveWindow(hMainWnd, rcMain.left, rcMain.top, rcMain.right, rcMain.bottom, TRUE);

  //Show main window menu
  SetMenu(hMainWnd, hMainMenu);

  if (!bMDI)
  {
    //Show main window children
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (hWndChild != hWndEdit)
        ShowWindow(hWndChild, SW_SHOW);
      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

    //Show edit window scrollbars
    if (bHideEditScrollBars)
    {
      ShowEditScrollBars(hWndEdit, TRUE);
    }
  }
  else
  {
    //Remove subclass
    if (NewFrameProcRetData)
    {
      SendMessage(hMainWnd, AKD_SETFRAMEPROCRET, (WPARAM)NULL, (LPARAM)&NewFrameProcRetData);
      NewFrameProcRetData=NULL;
    }

    //Show main window children
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (hWndChild == hStatus)
      {
        if (bStatusVisible)
          ShowWindow(hStatus, SW_SHOW);
      }
      else if (hWndChild == hTab)
      {
        if (bTabVisible)
          ShowWindow(hTab, SW_SHOW);
      }
      else if (hWndChild != hMdiClient)
        ShowWindow(hWndChild, SW_SHOW);

      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

    //Restore MDI frame windows
    if (hWndFrameActive=(HWND)SendMessage(hMdiClient, WM_MDIGETACTIVE, 0, (LPARAM)NULL))
    {
      if (bFrameMaximized)
      {
        SendMessage(hMdiClient, WM_SETREDRAW, FALSE, 0);
        StackShowFrameWindow(&hFrameWindowStack);
        SendMessage(hMdiClient, WM_MDIMAXIMIZE, (WPARAM)hWndFrameActive, 0);
        SendMessage(hMdiClient, WM_SETREDRAW, TRUE, 0);
        InvalidateRect(hWndFrameActive, NULL, FALSE);
      }
      else
      {
        StackShowFrameWindow(&hFrameWindowStack);
        SendMessage(hMdiClient, WM_MDIACTIVATE, (WPARAM)hWndFrameActive, 0);
      }
    }
  }

  //Focus on main window
  SetFocus(hMainWnd);

  StackFreeFrameWindow(&hFrameWindowStack);
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
